/**	$Revision: 1.18 $
	Last updated: $Date: 2002/09/17 16:13:01 $

	Copyright (C) 2000-2002 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "AccountsDB.h"

#include <DateTime.h>
#include <FeatureMgr.h>

#include "TransactionsDB.h"
#include "SplitsDB.h"
#include "AccountsTable.h"
#include "CoinsPref.h"

#include "FreeCoinsRsc.h"

AccountsDB*	accounts_db;
///the number of masks defined
static const UInt8		type_masks_no = 8;
///type masks for account type selector
static UInt32			type_masks[type_masks_no] = {0xffff, 3, 12, 1, 2,  4, 8, 16};

Int32	AccountsDB::getBalance(UInt32 uid)
{
	accountStructure	as;	
	
	getUnpackedRecord(uid, &as);
	return as.balance;
}

Int32	AccountsDB::getClearedBalance(UInt32 uid)
{
	accountStructure		 	as;	
	
	getUnpackedRecord(uid, &as);
	return as.cleared_balance;
}

void AccountsDB::unpackRecord (accountStructure* account, packedAccountStructure* packedAccount)
{
	Char*	s = packedAccount->description;
	
	account->balance = packedAccount->balance;
	account->cleared_balance = packedAccount->cleared_balance;
	account->credit_limit = packedAccount->credit_limit;
	account->parent_uid = packedAccount->parent_uid;
	account->children_total = packedAccount->children_total;
	account->children_cleared_total = packedAccount->children_cleared_total;
	account->children_available_total = packedAccount->children_available_total;
	account->type = (AccountType)packedAccount->type;
	account->depth = packedAccount->depth;
	account->children_no = packedAccount->children_no;
	account->collapsed = packedAccount->flags.collapsed;	
	account->credit_account = packedAccount->flags.credit_account;
	account->currency = packedAccount->currency;
			
	account->description = s;
	s += StrLen(s)+1;
	account->name = s;
	s += StrLen(s)+1;
	account->note = s;
}

void AccountsDB::packRecord (accountStructure* account, MemHandle accountDBEntry)
{
	// figure out necessary size
	UInt32	length = 0;
	MemPtr	s;
	UInt32	offset = 0;
	AccountFlags flags;
	
	//total length is the size of the packed struct + size of the strings
	length = sizeof(packedAccountStructure)
			+ StrLen(account->description) + StrLen(account->name) + StrLen (account->note) + 2;

	flags.collapsed = account->collapsed;
	flags.credit_account = account->credit_account;
	flags.reserved = 0;

	if (MemHandleResize(accountDBEntry, length) == 0) {
		// copy the fields
		s = MemHandleLock(accountDBEntry);
		writeAndAdvance(s,offset,&account->balance);
		writeAndAdvance(s,offset,&account->cleared_balance);
		writeAndAdvance(s,offset,&account->credit_limit);
		writeAndAdvance(s,offset,&account->parent_uid);
		writeAndAdvance(s,offset,&account->currency);
		writeAndAdvance(s,offset,&account->children_total);
		writeAndAdvance(s,offset,&account->children_cleared_total);
		writeAndAdvance(s,offset,&account->children_available_total);
		writeAndAdvance(s,offset,&flags);
		writeAndAdvance(s,offset,&account->type);
		writeAndAdvance(s,offset,&account->depth);
		writeAndAdvance(s,offset,&account->children_no);
		writeAndAdvance(s,offset,account->description);
		writeAndAdvance(s,offset,account->name);
		writeAndAdvance(s,offset,account->note);
	
		MemHandleUnlock(accountDBEntry);
	} else {
		ErrFatalDisplay ("resizing failed");
	}
}

AccountsDB::AccountsDB() : Database <accountStructure, packedAccountStructure> (ACCOUNTSDB_TYPE, FREE_COINS_CREATOR, ACCOUNTSDB_NAME, FREE_COINS_CATEGORIES)
{
	totals.total=totals.cleared_total=totals.available_total=0;
	current_cat=CategoryFind(ref,"Visible");
}

AccountsDB::AccountsDB(DmOpenRef db_open_ref) : Database <accountStructure, packedAccountStructure> (db_open_ref)
{
	totals.total=totals.cleared_total=totals.available_total=0;
	current_cat=CategoryFind(ref,"Visible");
}

UInt32 AccountsDB::createAccount(char *name, UInt32 parent_uid,
	 AccountType type)
{
	accountStructure as;
	UInt16 index;
	UInt32 uid;

	MemSet(&as, sizeof(accountStructure), 0);
	as.parent_uid = parent_uid;
	if (parent_uid) as.depth = 1;
	as.name = name;
	as.description = "";
	as.type = type;
	index = newRecord(&as);
	uid = getRecordUid(index);
	return uid;
}

void AccountsDB::incrementBalance(UInt32 uid, Int32 clr_amt, Int32 amount, Boolean is_parent /*= false*/)
{
	accountStructure	a;
	MemHandle			h;
	UInt32 type_mask;

	h = getUnpackedRecord(uid,&a);

	// if the account is visible, update grand total
	if (preferences.get()->visible_acc_types >= type_masks_no) {
		type_mask = 0xffff;
	} else {
		type_mask = type_masks[preferences.get()->visible_acc_types];
	}
	if ((1L<<((UInt8)a.type)) & type_mask) {
		totals.total += amount;
		totals.cleared_total += clr_amt;
		totals.available_total += amount;
	}
	if (is_parent) {
		a.children_total += amount;
		a.children_cleared_total += clr_amt;
		a.children_available_total += amount;
	} else {
		a.balance += amount;
		a.cleared_balance += clr_amt;
	}
	packRecord(&a,h);

	while (a.depth!=0) {
		uid = a.parent_uid;
		h=getUnpackedRecord(uid,&a);
		a.children_total += amount;
		a.children_cleared_total += clr_amt;
		a.children_available_total += amount;
		packRecord(&a,h);
	}
}

void AccountsDB::updateBalance(UInt32 uid,Int32 newbal, char* payee /*= NULL*/)
{
	accountStructure		a;
	transactionStructure	t;
	splitStructure			s;
	UInt16 index;
	MemHandle				handle=NULL; //handle for string resources

	getUnpackedRecord(uid,&a);

	if (a.balance == newbal)
		return;

	if (payee) t.payee = payee;
	else {
		handle=DmGetResource (strRsc,BALANCE_UPDATE_STR);
		t.payee=reinterpret_cast<char*> (MemHandleLock (handle));
	}

	DateSecondsToDate(TimGetSeconds(),& t.date);
	t.method=pm_transfer;
	t.num="";	
	t.note="";	
	t.num_splits = 1;
	index = transactions_db->newRecord(&t,true);	//cleared by default

	//unlock the acquired resource if necessary
	if(handle) {
		MemHandleUnlock(handle);   
		DmReleaseResource(handle); 
	}
	
	s.trans_uid = transactions_db->getRecordUid(index);
	s.acc_uid = uid;
	s.amount = newbal-a.balance;
	s.line_no = 1;
	splits_db->newRecord(&s, true);
}	


UInt32	AccountsDB::findAccountByName(char* n, UInt16 categ)
{
	UInt16				recordNo;
	UInt32				uniqueID;
	accountStructure	as;

  	for(recordNo=0;;recordNo++) {
		if(!DmQueryNextInCategory(ref,&recordNo,categ)) return 0;		
		DmRecordInfo(ref,recordNo,NULL,&uniqueID,NULL);

		getUnpackedRecord(uniqueID,&as);

		if(!StrCaselessCompare(n,as.name)) return uniqueID;
	}	
}

void AccountsDB::getFullNameTruncLeftToFit(UInt32 acc, Int16 width,
	Char *dest)
{
	UInt32 uids[10];
	accountStructure a;
	Char* s;
	Int16 l, i;
	Char ellipsis;

	l = 0;
	i = 0;
	do {
		getUnpackedRecord(acc, &a);
		l += StrLen(a.name);
		uids[i] = acc;
		acc = a.parent_uid;
		i++;
	} while (a.depth>0 && i<10);
	l += i;
	i -= 2;
	s = (Char*)MemPtrNew(l);
	StrCopy(s, a.name);
	while(i>=0) {
		getUnpackedRecord(uids[i], &a);
		StrCat(s,":");
		StrCat(s, a.name); 
		i--;
	}

	for (i=l-1; i>=0; i--) {
		width -= FntCharWidth(s[i]);
		if (width<0) break;
	}
	if (width<0) {
		ChrHorizEllipsis(&ellipsis);
		width -= FntCharWidth(ellipsis);
		while (width<0) {
			width += FntCharWidth(s[i]);
			i++;
		}
		s[i] = ellipsis;
		MemMove(s, s+i, l-i);
	}
	StrCopy(dest, s);
	MemPtrFree(s);
}

void	AccountsDB::getAccountName(UInt32 acc, Char* buf)
{
	accountStructure as;
	
	getUnpackedRecord(acc,&as);
	StrCopy(buf,as.name);
}

UInt32	AccountsDB::getAccountCurrency(UInt32 acc)
{
	accountStructure as;
	
	getUnpackedRecord(acc,&as);
	return as.currency;
}

Boolean	AccountsDB::isDisplayable(UInt16 acc)
{
	accountStructure as;
	
	getUnpackedRecord(acc,&as);
	if(as.depth == 0) return true; //all root accounts are visible
	for(UInt32 parent = as.parent_uid; as.depth!=0; parent = as.parent_uid)
	{
		getUnpackedRecord(parent,&as);
		if(as.collapsed) {
			return false;
		}
	}
	return true;
}


UInt16	AccountsDB::newRecord(accountStructure* s)
{
	accountStructure	parent;
	MemHandle		h,p,parent_h;
	UInt16			index;
	Int32			bal;
	
	//clear the balance since it is maintained via the transaction
	bal=s->balance;
	s->balance=s->cleared_balance=0;
	s->collapsed=true;
	if(s->depth!=0) { //update parent information
		parent_h=getUnpackedRecord(s->parent_uid,&parent);
		parent.children_no++;
		//overwrite the depth
		s->depth=parent.depth+1;		
		packRecord(&parent,parent_h);

		MemHandle tmp_h;
		accountStructure a;
		a.depth = s->depth;
		a.parent_uid = s->parent_uid;
		while (a.depth!=0) {
			tmp_h=getUnpackedRecord(a.parent_uid,&a);
			a.children_available_total += s->credit_limit;
			packRecord(&a,tmp_h);
		}
	}

	//set size to 1 - will be resized by packrecord
	h=DmNewHandle(ref,1);
	
	if(!h) {
		MemHandle str_handle = DmGetResource (strRsc,NEW_REC_FAILED_STR);
		ErrFatalDisplay(reinterpret_cast<char*> (MemHandleLock (str_handle)));
		MemHandleUnlock (str_handle);
		DmReleaseResource (str_handle);
	}

	packRecord(s,h);

	p = (MemHandle) MemHandleLock(h);
	index=DmFindSortPosition(ref,p, 0, (DmComparF *) compareRecords, 0);
	MemPtrUnlock(p);
	
	DmAttachRecord(ref,&index,h,0); //automatically sets the dirty bit

	if(current_cat!=dmAllCategories) setRecordCategory(index,current_cat);

	if (bal!=0) {
		UInt32 acc_uid;
		DmRecordInfo(ref,index,NULL,&acc_uid,NULL);

		MemHandle str_handle = DmGetResource (strRsc,ACC_OPEN_STR);
		updateBalance(acc_uid, bal,reinterpret_cast<char*> (MemHandleLock (str_handle)) );
		MemHandleUnlock (str_handle);
		DmReleaseResource (str_handle);
	}
	
	return index;	
}

bool	AccountsDB::deleteRecord(UInt16	id)
{
	UInt16				recordNo;
	UInt32				trans_uid,acc_uid;
	UInt16				index, count;
	transactionStructure	ts;
	accountStructure		a;
	splitStructure		s;
	MemHandle			h;
 	
	//sanity check: if deleting the account set as default, unset the default account
	if(getRecordUid(id) == preferences.get()->def_acc) {
		preferences.get()->def_acc = 0;
	}
	
	acc_uid = getRecordUid(id);

	// if necessary, notify parent that we're gone
	getUnpackedRecord(acc_uid, &a);
	if (a.depth!=0) {
		UInt32 p_uid = a.parent_uid;
		h = getUnpackedRecord(p_uid,&a);
		a.children_no--;
		packRecord(&a,h);
	}

	// in all transactions delete all splits that mention this account
	for(recordNo=0;;recordNo++) {
		if(!DmQueryNextInCategory(transactions_db->getRef(),&recordNo,dmAllCategories)) break;				
		DmRecordInfo(transactions_db->getRef(),recordNo,NULL,&trans_uid,NULL);
		h = transactions_db->getUnpackedRecord(trans_uid,&ts);

		count = ts.num_splits;
		for (index=splits_db->locateFirstSplit(trans_uid);count>0;
			index++,count--){
			splits_db->getUnpackedRecord(index, &s);
			if (s.acc_uid==acc_uid) {
				splits_db->deleteRecord(index);
				index--;
				ts.num_splits--;
			}
		}
		transactions_db->packRecord(&ts,h);
		if (ts.num_splits==0) {
			transactions_db->deleteRecord(recordNo);
			recordNo--;
		}
	}
	//let the database class delete the record
	return Database <accountStructure, packedAccountStructure>::deleteRecord(id);
}


Int16	AccountsDB::compareRecords(packedAccountStructure* rec1,packedAccountStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr sr2,MemHandle app_info)
{
	accountStructure ac1,ac2;
	accounts_db->unpackRecord(&ac1,rec1);
	accounts_db->unpackRecord(&ac2,rec2);
	Int8 result = 0, tie_breaker = 0;

	// if the procedure below results in a tie, then the two accounts were
	// on the same branch. In that case, sort the higher one first, the deeper
    // one second.
	tie_breaker = ac1.depth - ac2.depth;

	// trace the deeper node back to the same level as the higher node
	while (ac1.depth != ac2.depth) {
		if (ac1.depth > ac2.depth) {
			accounts_db->getUnpackedRecord(ac1.parent_uid,&ac1);
		}
		else {
			accounts_db->getUnpackedRecord(ac2.parent_uid,&ac2);
		}
	}

	// trace both back until they have the same parent
    while (ac1.depth!=0 && ac1.parent_uid != ac2.parent_uid) {
		accounts_db->getUnpackedRecord(ac1.parent_uid,&ac1);
		accounts_db->getUnpackedRecord(ac2.parent_uid,&ac2);
	}

	// now they're comparable.
	result = StrCaselessCompare(ac1.name, ac2.name);

	if (result==0)
		result = tie_breaker;
	return result;
}

void AccountsDB::updateVisibility()
{
	accountStructure	as;
	UInt32 			type_mask = 0;
	UInt16			recordNo;
	UInt16 			visible = CategoryFind(ref, "Visible");
	UInt16 			invisible = CategoryFind(ref, "Invisible");

	if (preferences.get()->visible_acc_types >= type_masks_no) {
		type_mask = 0xffff;
	}
	else {
		type_mask = type_masks[preferences.get()->visible_acc_types];
	}

	totals.total = totals.cleared_total = totals.available_total= 0;
  	for(recordNo=0;;recordNo++) {
		if(!DmQueryNextInCategory(ref,&recordNo,dmAllCategories))
			break;
		getUnpackedRecord(recordNo,&as);
		if ((1L<<((UInt8)as.type)) & type_mask) {
			setRecordCategory(recordNo, visible);
			totals.total += as.balance;
			totals.cleared_total += as.cleared_balance;
			totals.available_total += as.balance + as.credit_limit;
		} else {
			setRecordCategory(recordNo, invisible);
		}
	}
}
